@@ -24,35 +24,53 @@
    	Pavel Cisler <pavel@eazel.com>
  */
 
-#ifdef HAVE_CONFIG_H
 #include <config.h>
-#endif
 
-#include <glib.h>
-#include <libgnome/gnome-defs.h>
-#include <libgnome/gnome-i18n.h>
+#define _LARGEFILE64_SOURCE
+
+#include <libgnomevfs/gnome-vfs-cancellation.h>
+#include <libgnomevfs/gnome-vfs-context.h>
+#include <libgnomevfs/gnome-vfs-i18n.h>
+#include <libgnomevfs/gnome-vfs-method.h>
+#include <libgnomevfs/gnome-vfs-mime.h>
+#include <libgnomevfs/gnome-vfs-module-shared.h>
+#include <libgnomevfs/gnome-vfs-module.h>
+#include <libgnomevfs/gnome-vfs-utils.h>
+#include <libgnomevfs/gnome-vfs-mime.h>
+#include <libgnomevfs/gnome-vfs-monitor-private.h>
 #include <dirent.h>
 #include <errno.h>
 #include <fcntl.h>
+#include <glib/gstrfuncs.h>
+#include <glib/gutils.h>
 #include <limits.h>
-#include <malloc.h>
+#include <stdio.h>
+#include <stdlib.h>
+#include <string.h>
 #include <sys/stat.h>
 #include <sys/types.h>
 #include <unistd.h>
-#include <stdio.h>
 #include <utime.h>
 #include <string.h>
+#ifdef HAVE_FAM
+#include <fam.h>
+#include <glib/giochannel.h>
+#endif
 
-#include "gnome-vfs-mime.h"
 
-#include "gnome-vfs-cancellation.h"
-#include "gnome-vfs-context.h"
-#include "gnome-vfs-module.h"
-#include "gnome-vfs-method.h"
-#include "gnome-vfs-utils.h"
-#include "gnome-vfs-module-shared.h"
-#include "file-method.h"
 
+/*#include "file-method.h"*/
+
+#ifdef HAVE_FAM
+FAMConnection *fam_connection = NULL;
+G_LOCK_DEFINE_STATIC (fam_connection);
+
+typedef struct {
+	FAMRequest request;
+	GnomeVFSURI *uri;
+} FileMonitorHandle;
+
+#endif
 
 #ifdef PATH_MAX
 #define	GET_PATH_MAX()	PATH_MAX
@@ -482,15 +500,12 @@ typedef struct {
 
 	gchar *name_buffer;
 	gchar *name_ptr;
-
-	const GnomeVFSDirectoryFilter *filter;
 } DirectoryHandle;
 
 static DirectoryHandle *
 directory_handle_new (GnomeVFSURI *uri,
 		      DIR *dir,
-		      GnomeVFSFileInfoOptions options,
-		      const GnomeVFSDirectoryFilter *filter)
+		      GnomeVFSFileInfoOptions options)
 {
 	DirectoryHandle *result;
 	gchar *full_name;
@@ -519,7 +534,6 @@ directory_handle_new (GnomeVFSURI *uri,
 	g_free (full_name);
 
 	result->options = options;
-	result->filter = filter;
 
 	return result;
 }
@@ -703,7 +717,6 @@ do_open_directory (GnomeVFSMethod *method,
 		   GnomeVFSMethodHandle **method_handle,
 		   GnomeVFSURI *uri,
 		   GnomeVFSFileInfoOptions options,
-		   const GnomeVFSDirectoryFilter *filter,
 		   GnomeVFSContext *context)
 {
 	gchar *directory_name;
@@ -720,8 +733,7 @@ do_open_directory (GnomeVFSMethod *method,
 
 	*method_handle
 		= (GnomeVFSMethodHandle *) directory_handle_new (uri, dir,
-								 options,
-								 filter);
+								 options);
 
 	return GNOME_VFS_OK;
 }
@@ -742,31 +754,19 @@ do_close_directory (GnomeVFSMethod *method,
 	return GNOME_VFS_OK;
 }
 
-inline static GnomeVFSResult
-read_directory (DirectoryHandle *handle,
-		GnomeVFSFileInfo *info,
-		gboolean *skip,
-		GnomeVFSContext *context)
+static GnomeVFSResult
+do_read_directory (GnomeVFSMethod *method,
+		   GnomeVFSMethodHandle *method_handle,
+		   GnomeVFSFileInfo *file_info,
+		   GnomeVFSContext *context)
 {
-	const GnomeVFSDirectoryFilter *filter;
-	GnomeVFSDirectoryFilterNeeds filter_needs;
 	struct dirent *result;
 	struct stat statbuf;
 	gchar *full_name;
-	gboolean filter_called;
-
-	/* This makes sure we don't try to filter the file more than
-           once.  */
-	filter_called = FALSE;
-	*skip = FALSE;
-
-	filter = handle->filter;
-	if (filter != NULL) {
-		filter_needs = gnome_vfs_directory_filter_get_needs (filter);
-	} else {
-		filter_needs = GNOME_VFS_DIRECTORY_FILTER_NEEDS_NOTHING;
-	}
+	DirectoryHandle *handle;
 
+	handle = (DirectoryHandle *) method_handle;
+	
 	if (readdir_r (handle->dir, handle->current_entry, &result) != 0) {
 		return gnome_vfs_result_from_errno ();
 	}
@@ -775,26 +775,12 @@ read_directory (DirectoryHandle *handle,
 		return GNOME_VFS_ERROR_EOF;
 	}
 
-	info->name = g_strdup (result->d_name);
-
-	if (filter != NULL
-	    && !filter_called
-	    && (filter_needs
-		  & (GNOME_VFS_DIRECTORY_FILTER_NEEDS_TYPE
-		     | GNOME_VFS_DIRECTORY_FILTER_NEEDS_STAT
-		     | GNOME_VFS_DIRECTORY_FILTER_NEEDS_MIMETYPE)) == 0){
-		if (!gnome_vfs_directory_filter_apply (filter, info)) {
-			*skip = TRUE;
-			return GNOME_VFS_OK;
-		}
-
-		filter_called = TRUE;
-	}
+	file_info->name = g_strdup (result->d_name);
 
 	strcpy (handle->name_ptr, result->d_name);
 	full_name = handle->name_buffer;
 
-	if (get_stat_info (info, full_name, handle->options, &statbuf) != GNOME_VFS_OK) {
+	if (get_stat_info (file_info, full_name, handle->options, &statbuf) != GNOME_VFS_OK) {
 		/* Return OK - this should not terminate the directory iteration
 		 * and we will know from the valid_fields that we don't have the
 		 * stat info.
@@ -802,60 +788,13 @@ read_directory (DirectoryHandle *handle,
 		return GNOME_VFS_OK;
 	}
 	
-	if (filter != NULL && !filter_called
-	    && (filter_needs & GNOME_VFS_DIRECTORY_FILTER_NEEDS_MIMETYPE) == 0) {
-		if (!gnome_vfs_directory_filter_apply (filter, info)) {
-			*skip = TRUE;
-			return GNOME_VFS_OK;
-		}
-		filter_called = TRUE;
-	}
-
 	if (handle->options & GNOME_VFS_FILE_INFO_GET_MIME_TYPE) {
-		get_mime_type (info, full_name, handle->options, &statbuf);
-	}
-
-	if (filter != NULL && !filter_called) {
-		if (!gnome_vfs_directory_filter_apply (filter, info)) {
-			*skip = TRUE;
-			return GNOME_VFS_OK;
-		}
-		filter_called = TRUE;
-	}
-
-	if (filter != NULL && !filter_called) {
-		if (!gnome_vfs_directory_filter_apply (filter, info)) {
-			*skip = TRUE;
-			return GNOME_VFS_OK;
-		}
-		filter_called = TRUE;
+		get_mime_type (file_info, full_name, handle->options, &statbuf);
 	}
 
 	return GNOME_VFS_OK;
 }
 
-static GnomeVFSResult
-do_read_directory (GnomeVFSMethod *method,
-		   GnomeVFSMethodHandle *method_handle,
-		   GnomeVFSFileInfo *file_info,
-		   GnomeVFSContext *context)
-{
-	GnomeVFSResult result;
-	gboolean skip;
-
-	do {
-		result = read_directory ((DirectoryHandle *) method_handle,
-					 file_info, &skip, context);
-		if (result != GNOME_VFS_OK)
-			break;
-		if (skip)
-			gnome_vfs_file_info_clear (file_info);
-	} while (skip);
-
-	return result;
-}
-
-
 static GnomeVFSResult
 do_get_file_info (GnomeVFSMethod *method,
 		  GnomeVFSURI *uri,
@@ -2052,6 +1991,170 @@ do_set_file_info (GnomeVFSMethod *method,
 	return GNOME_VFS_OK;
 }
 
+#ifdef HAVE_FAM
+static gboolean
+fam_callback (GIOChannel *source,
+	      GIOCondition condition,
+	      gpointer data)
+{
+	FileMonitorHandle *handle;
+	GnomeVFSURI *info_uri = NULL;
+	GnomeVFSMonitorEventType event_type = GNOME_VFS_MONITOR_EVENT_CHANGED;
+
+	G_LOCK (fam_connection);
+
+	while (FAMPending(fam_connection)) {
+		FAMEvent ev;
+		if (FAMNextEvent(fam_connection, &ev) != 1) {
+			FAMClose(fam_connection);
+			g_free(fam_connection);
+			fam_connection = FALSE;
+			G_UNLOCK (fam_connection);
+			return FALSE;
+		}
+
+		handle = (FileMonitorHandle *)ev.userdata;
+
+		info_uri = gnome_vfs_uri_append_file_name (handle->uri, 
+				ev.filename);
+
+		switch (ev.code) {
+			case FAMChanged:
+				event_type = GNOME_VFS_MONITOR_EVENT_CHANGED;
+				break;
+			case FAMDeleted:
+				event_type = GNOME_VFS_MONITOR_EVENT_DELETED;
+				break;
+			case FAMStartExecuting:
+				event_type = 
+					GNOME_VFS_MONITOR_EVENT_STARTEXECUTING;
+				break;
+			case FAMStopExecuting:
+				event_type = 
+					GNOME_VFS_MONITOR_EVENT_STOPEXECUTING;
+				break;
+			case FAMCreated:
+				event_type = GNOME_VFS_MONITOR_EVENT_CREATED;
+				break;
+			case FAMAcknowledge:
+			case FAMExists:
+			case FAMEndExist:
+			case FAMMoved:
+				event_type = -1;
+				g_warning ("bad FAM event: %d", ev.code);
+				break;
+		}
+
+
+		if (event_type != -1) {
+			gnome_vfs_monitor_callback (
+					(GnomeVFSMethodHandle *)handle,
+					info_uri, event_type);
+		}
+	}
+
+	G_UNLOCK (fam_connection);
+
+	return TRUE;
+}
+
+
+
+static gboolean
+monitor_setup (void)
+{
+	GIOChannel *ioc;
+	gint watch_id;
+
+	G_LOCK (fam_connection);
+
+	if (fam_connection == NULL) {
+		fam_connection = g_malloc0(sizeof(FAMConnection));
+		if (FAMOpen2(fam_connection, "test-monitor") != 0) {
+			g_print ("FAMOpen failed, FAMErrno=%d\n", FAMErrno);
+			g_free(fam_connection);
+			fam_connection = NULL;
+			return FALSE;
+		}
+		ioc = g_io_channel_unix_new (FAMCONNECTION_GETFD(fam_connection));
+		watch_id = g_io_add_watch (ioc,
+					   G_IO_IN | G_IO_HUP | G_IO_ERR,
+					   fam_callback, fam_connection);
+		g_io_channel_unref (ioc);
+	}
+
+	G_UNLOCK (fam_connection);
+
+	return TRUE;
+}
+#endif
+
+static GnomeVFSResult
+do_monitor_add (GnomeVFSMethod *method,
+		GnomeVFSMethodHandle **method_handle_return,
+		GnomeVFSURI *uri,
+		GnomeVFSMonitorType monitor_type)
+{
+#ifdef HAVE_FAM
+	FileMonitorHandle *handle;
+	char *filename;
+
+	if (!monitor_setup ()) {
+		return GNOME_VFS_ERROR_NOT_SUPPORTED;
+	}
+
+	handle = g_new0 (FileMonitorHandle, 1);
+	handle->uri = uri;
+	gnome_vfs_uri_ref (uri);
+	filename = get_path_from_uri (uri);
+	
+	if (monitor_type == GNOME_VFS_MONITOR_FILE) {
+		G_LOCK (fam_connection);
+		FAMMonitorFile (fam_connection, filename, 
+			&handle->request, handle);
+		G_UNLOCK (fam_connection);
+	} else {
+		G_LOCK (fam_connection);
+		FAMMonitorDirectory (fam_connection, filename, 
+			&handle->request, handle);
+		G_UNLOCK (fam_connection);
+	}
+
+	*method_handle_return = (GnomeVFSMethodHandle *)handle;
+
+	g_free (filename);
+
+	return GNOME_VFS_OK;
+#else
+	return GNOME_VFS_ERROR_NOT_SUPPORTED;
+#endif
+}
+
+static GnomeVFSResult
+do_monitor_cancel (GnomeVFSMethod *method,
+		   GnomeVFSMethodHandle *method_handle)
+{
+#ifdef HAVE_FAM
+	FileMonitorHandle *handle = (FileMonitorHandle *)method_handle;
+
+	if (!monitor_setup ()) {
+		return GNOME_VFS_ERROR_NOT_SUPPORTED;
+	}
+
+	G_LOCK (fam_connection);
+	FAMCancelMonitor (fam_connection, &handle->request);
+	G_UNLOCK (fam_connection);
+
+	gnome_vfs_uri_unref (handle->uri);
+
+	g_free (handle);
+
+	return GNOME_VFS_OK;
+#else
+	return GNOME_VFS_ERROR_NOT_SUPPORTED;
+#endif
+}
+
 static GnomeVFSMethod method = {
 	sizeof (GnomeVFSMethod),
 	do_open,
@@ -2076,7 +2179,9 @@ static GnomeVFSMethod method = {
 	do_set_file_info,
 	do_truncate,
 	do_find_directory,
-	do_create_symbolic_link
+	do_create_symbolic_link,
+	do_monitor_add,
+	do_monitor_cancel
 };
 
 GnomeVFSMethod *
